"""
The command line interface to cfn_nagger.

"""
from __future__ import absolute_import, division, print_function
import sys
import inspect
import logging
import traceback
from configparser import RawConfigParser
import boto3
import click
from aws_bucket_creator import BucketCreator
import aws_bucket_creator

def lineno():
    """Returns the current line number in our program."""
    return str(' - BucketCreator - line number: '+str(inspect.currentframe().f_back.f_lineno))


@click.group()
@click.version_option(version='0.0.16')
def cli():
    pass


@cli.command()
@click.option('--ini', '-i', help='INI file with needed information', required=True)
@click.option('--version', '-v', help='Print version and exit', required=False, is_flag=True)
@click.option('--debug', help='Turn on debugging', required=False, is_flag=True)
def upsert(
        ini,
        version,
        debug
    ):
    '''
    primary function for creating a bucket
    :return:
    '''

    ini_data = read_config_info(ini)
    if 'environment' not in ini_data:
        print('[environment] section is required in the INI file')
        sys.exit(1)


    if debug:
        ini_data['parameters']['debug'] = True
    else:
        ini_data['parameters']['debug'] = False


    if 'region' not in ini_data['environment']:
        ini_data['environment']['region'] = find_myself()

    if version:
        myversion()
    else:
        start_create(
            ini_data,
            debug
        )




@click.option('--version', '-v', help='Print version and exit', required=False, is_flag=True)
def version(version):
    """
    Get version
    :param version:
    :return:
    """
    myversion()


def myversion():
    '''
    Gets the current version
    :return: current version
    '''
    print('Version: '+str(aws_bucket_creator.__version__))

def start_create(
        ini,
        debug
    ):
    '''
    Starts the creation
    :return:
    '''
    if debug:
        print('command - start_create'+lineno())
        print('ini data: '+str(ini)+lineno())


    config_dict = {}

    if 'debug' in ini['parameters']:

        if ini['parameters']['debug']:
            config_dict['debug'] = ini['parameters']['debug']
        else:
            config_dict['debug'] = False
    else:
        config_dict['debug'] = False

    if ini['parameters']['bucket_name']:
        config_dict['bucket_name'] = ini['parameters']['bucket_name']
    if ini['environment']['profile']:
        config_dict['aws_profile'] = ini['environment']['profile']
    if ini['environment']['region']:
        config_dict['region'] = ini['environment']['region']
    tags = []

    found_name_tag = -1
    for tag in ini['tags']:

        if str(tag) == 'Name' or str(tag) == 'name':
            found_name_tag = 1
        temp_dict = {}
        temp_dict['Key'] = tag
        temp_dict['Value'] = ini['tags'][tag]
        tags.append(temp_dict)

    # Add bucket name as tag
    if found_name_tag < 0:
        temp_dict = {}
        temp_dict['Key'] = 'Name'
        temp_dict['Value'] = ini['parameters']['bucket_name']
        tags.append(temp_dict)


    if tags:
        config_dict['tags'] = tags

    if 'logging_enable' in ini['parameters']:
        config_dict['logging_enabled'] = ini['parameters']['logging_enabled']
    else:
        config_dict['logging_enabled'] = False

    if 'public_write_access' in ini['parameters']:
        config_dict['public_write_access'] = ini['parameters']['public_write_access']

    if 'acl' in ini['parameters']:
        config_dict['acl'] = ini['parameters']['acl']
    else:
        config_dict['acl'] = 'bucket-owner-full-control'

    if 'days_to_glacier' in ini['parameters']:
        config_dict['days_to_glacier'] = ini['parameters']['days_to_glacier']
    else:
        config_dict['days_to_glacier'] = 365

    if 'days_to_standard_ia' in ini['parameters']:
        config_dict['days_to_standard_ia'] = ini['parameters']['days_to_standard_ia']
    else:
        config_dict['days_to_standard_ia'] = 30


    if 'principals' in ini['parameters']:
        config_dict['bucket_policy_principals'] = ini['parameters']['principals']

    if 'bucket_policy' in ini['parameters']:
        config_dict['bucket_policy'] = ini['parameters']['bucket_policy']


    if 'bucket_policy_path' in ini['parameters']:
        config_dict['days_to_standard_ia'] = ini['parameters']['bucket_policy_path']
    else:
        config_dict['bucket_policy_path'] = None

    if 'event_lambda_arn' in ini['parameters']:
        config_dict['event_lambda_arn'] = ini['parameters']['event_lambda_arn']


    creator = BucketCreator(config_dict)
    if debug:
        print('print have BucketCreator')
    if creator.create():
        if debug:
            print('created')
    else:
        if debug:
            print('not created')

def find_myself():
    """
    Find myself
    Args:
        None
    Returns:
       An Amazon region
    """
    my_session = boto3.session.Session()
    return my_session.region_name

def read_config_info(ini_file):
    """
    Read the INI file
    Args:
        ini_file - path to the file
    Returns:
        A dictionary of stuff from the INI file
    Exits:
        1 - if problems are encountered
    """
    try:
        config = RawConfigParser()
        config.optionxform = lambda option: option
        config.read(ini_file)
        the_stuff = {}
        for section in config.sections():
            the_stuff[str(section)] = {}
            for option in config.options(section):
                the_stuff[str(section)][str(option)] = str(config.get(section, option.replace('\n', '')))

        return the_stuff
    except Exception as wtf:
        logging.error('Exception caught in read_config_info(): {}'.format(wtf))
        traceback.print_exc(file=sys.stdout)
        return sys.exit(1)






from __future__ import absolute_import, division, print_function
import pkg_resources
from aws_bucket_creator.BucketCreator import BucketCreator #noqa

__version__ = pkg_resources.get_distribution('aws_bucket_creator').version

__all__ = [
]
__title__ = 'aws_bucket_creator'
__version__ = '0.0.16'
__author__ = 'Will Rubel'
__author_email__ = 'willrubel@gmail.com'




from __future__ import absolute_import, division, print_function
import logging
import inspect
import os
import sys
import json
import time
import botocore
import boto3
from botocore.exceptions import ClientError
from boto.s3.acl import ACL, Grant



def lineno():
    """Returns the current line number in our program."""
    return str(' - BucketCreator - line number: '+str(inspect.currentframe().f_back.f_lineno))


class BucketCreator:
    """
    Creates an S3 Bucket
    """

    def __init__(self, config_block):
        """
        Initialize BucketCreator
        :param config_block:
        """

        self.debug = False
        self.bucket_name = None
        self.aws_profile = None
        self.session = None
        self.client = None
        self.resource = None
        self.tags = None
        self.bucket_policy_principals = []
        self.bucket_policy = None
        self.bucket_policy_path = None
        self.days_to_glacier = 365
        self.days_to_standard_ia = 30
        self.acl = None
        self.bucket_owner_id = None
        self.bucket_owner_display_name = None
        self.public_write_access = False
        self.region = None
        self.event_lambda_arn = None
        self.logging_enabled = False


        if config_block:
            self._config = config_block
        else:
            logging.error('config block was garbage')
            raise SystemError

        if 'debug' in self._config:
            self.debug = self._config['debug']


        if 'tags' in self._config:
            self.tags = self._config['tags']

        if 'logging_enabled' in self._config:
            self.logging_enabled = self._config['logging_enabled']


        if 'region' in self._config:
            self.region = self._config['region']

        if 'public_write_access' in self._config:
            self.public_write_access = self._config['public_write_access']

        if 'acl' in self._config:
            self.acl = self._config['acl']

        if 'days_to_glacier' in self._config:
            self.days_to_glacier = int(self._config['days_to_glacier'])

        if 'days_to_standard_ia' in self._config:
            self.days_to_standard_ia = int(self._config['days_to_standard_ia'])

        if 'bucket_policy_principals' in self._config:
            self.bucket_policy_principals = self._config['bucket_policy_principals'].split(',')

        if 'bucket_name' in self._config:
            self.bucket_name = self._config['bucket_name']

        if 'bucket_policy_path' in self._config:
            self.bucket_policy_path = self._config['bucket_policy_path']

        if 'event_lambda_arn' in self._config:
            self.event_lambda_arn = self._config['event_lambda_arn']

        if 'bucket_policy' in self._config:
            self.bucket_policy = self._config['bucket_policy']


            if self.debug:
                print('bucket policy is: '+str(self.bucket_policy)+lineno())

        # Get boto session
        if self.region:
            self.session = boto3.session.Session(
                profile_name=self._config['aws_profile'],
                region_name=self.region
            )

        elif self.aws_profile:
            self.session = boto3.session.Session(profile_name=self._config['aws_profile'])
        else:
            self.session = boto3.session.Session()

        self.client = self.session.client('s3')
        self.resource = self.session.resource('s3')

        if self.debug:
            print('s3 bucket: '+str(self.bucket_name))


    def create(self):
        """
        Create a bucket
        :return: rendered results
        """

        if self.debug:
            print('BucketCreator - create'+lineno())

        self.create_bucket()
        print('Bucket created')
        self.get_bucket_owner_id()
        print('Found bucket owner id')
        self.create_acl()
        print('Bucket acl created')
        self.create_encryption()
        print('Encryption set')
        self.create_logging()
        print('Bucket logging set')
        self.create_tags()
        print('Bucket tags set')

        if len(self.bucket_policy_principals) > 0 and not self.bucket_policy_path:

            if self.debug:
                print('There are principals in the policy but no bucket policy path'+lineno())

            if not self.bucket_policy:
                self.create_bucket_policy()
            else:
                if self.debug:
                    print('## There is a bucket policy being passed-in')

            self.add_bucket_policy()
        elif len(self.bucket_policy_principals) < 1 and self.bucket_policy and not self.bucket_policy_path:
            if self.debug:
                print('## There is a bucket policy being passed-in')

            self.add_bucket_policy()

        elif self.bucket_policy_path:
            if self.debug:
                print('There is a bucket policy path...creating bucket policy'+lineno())
            self.bucket_policy = self.load_bucket_policy()
            self.add_bucket_policy()
        self.add_lifecycle_policy()

        self.create_bucket_notifications()
        print('Bucket notifications set')


    def create_bucket_notifications(self):
        """
        Create bucket notification
        :return:
        """
        if self.event_lambda_arn:

            try:
                response = self.client.put_bucket_notification_configuration(
                    Bucket=self.bucket_name,
                    NotificationConfiguration={
                        'LambdaFunctionConfigurations': [
                            {
                                'Id': 'MyEvent',
                                'LambdaFunctionArn': self.event_lambda_arn,
                                'Events': [
                                    's3:ObjectCreated:Put',
                                ]
                            }
                        ]
                    }
                )

                if self.debug:
                    print('response: ' + str(response))

            except ClientError as err:
                print('Could not get create bucket event: ' + str(err))



    def get_bucket_owner_id(self):
        """
        Get bucket owner id
        :return:
        """
        try:
            response = self.client.get_bucket_acl(
                Bucket=self.bucket_name
            )

            if self.debug:
                print('response: '+str(response))

            if 'Owner' in response:
                self.bucket_owner_id = response['Owner']['ID']
                self.bucket_owner_display_name = response['Owner']['DisplayName']

        except ClientError as err:
            print('Could not get bucket owner id: '+str(err))



    def create_acl(self):
        """
        Create bucket ACL
        :return:
        """
        try:
            response = self.client.get_bucket_logging(
                Bucket=self.bucket_name
            )

            if self.debug:
                print('acl response: '+str(response))

            if self.public_write_access:
                response = self.client.put_bucket_acl(
                    Bucket=self.bucket_name,
                    AccessControlPolicy={
                        'Grants': [
                            {
                                'Grantee': {
                                    'Type': 'Group',
                                    'URI': 'http://acs.amazonaws.com/groups/global/AllUsers',
                                },
                                'Permission': 'WRITE'
                            },
                            {
                                'Grantee': {
                                    'Type': 'Group',
                                    'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                },
                                'Permission':  'WRITE'
                            },
                            {
                                'Grantee': {
                                    'Type': 'Group',
                                    'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                },
                                'Permission': 'READ_ACP'
                            },
                        ],
                        'Owner': {
                            'DisplayName': 'Owner',
                            'ID': self.bucket_owner_id
                        }
                    }
                )

                if self.debug:
                    print(response)

            else:
                response = self.client.put_bucket_acl(
                    Bucket=self.bucket_name,
                    AccessControlPolicy={
                        'Grants': [
                            {
                                'Grantee': {
                                    'Type': 'Group',
                                    'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                },
                                'Permission':  'WRITE'
                            },
                            {
                                'Grantee': {
                                    'Type': 'Group',
                                    'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                },
                                'Permission': 'READ_ACP'
                            },
                        ],
                        'Owner': {
                            'DisplayName': 'Owner',
                            'ID': self.bucket_owner_id
                        }
                    }
                )
            if self.debug:
                print(response)

        except ClientError as err:
            print('Error creating bucket acl: '+str(err))



    def load_bucket_policy(self):
        """
        Load a bucket policy from a file
        :return:
        """
        if self.debug:
            print('load bucket policy')
            print('current directory is: ' + str(os.getcwd()))

        if str(self.bucket_policy_path).startswith('./'):
            self.bucket_policy_path = str(os.getcwd())+'/'+str(self.bucket_policy_path.replace('./', ''))

            if self.debug:
                print('new bucket path is: '+str(self.bucket_policy_path))
        if not str(self.bucket_policy_path).startswith('/'):
            self.bucket_policy_path = os.getcwd()+'/'+str(self.bucket_policy_path)

            if self.debug:
                print('new bucket path is: '+str(self.bucket_policy_path))

        try:
            with open(self.bucket_policy_path, 'r') as tempfile:  # OSError if file exists or is invalid
                pass

            with open(self.bucket_policy_path) as policy_file:
                bucket_policy = json.dumps((json.load(policy_file)))

                if self.debug:
                    print('bucket policy is: ' + str(self.bucket_policy)+lineno())

                return bucket_policy

        except OSError:
            print('could not open the bucket policy file')



    def create_logging(self):
        """
        Turn on logging for a bucket, or turn it off
        :return:
        """

        if self.debug:
            print('######################')
            print('Create logging')
            print('flag: '+str(self.logging_enabled))
            print('######################')

        try:
            response = self.client.get_bucket_logging(
                Bucket=self.bucket_name
            )

            if self.debug:
                print('bucket logging response: '+str(response))


            if not 'LoggingEnabled' in response:

                if self.debug:
                    print('no bucket logging')

                if self.logging_enabled:

                    response = self.client.put_bucket_logging(
                        Bucket=self.bucket_name,
                        BucketLoggingStatus={
                            'LoggingEnabled': {
                                'TargetBucket': self.bucket_name,
                                'TargetPrefix': 'user/',
                                'TargetGrants': [
                                    {
                                        'Grantee': {
                                            'Type': 'Group',
                                            'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                        },
                                        'Permission': 'WRITE'
                                    },
                                    {
                                        'Grantee': {
                                            'Type': 'Group',
                                            'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
                                        },
                                        'Permission': 'READ'
                                    },
                                ]
                            }
                        }

                    )

                    if self.debug:
                        print(response)

            # If logging is enabled
            else:
                if not self.logging_enabled:

                    response = self.client.put_bucket_logging(
                        Bucket=self.bucket_name,
                        BucketLoggingStatus={
                        }

                    )

                    if self.debug:
                        print(response)

        except ClientError as err:
            print('Error creating bucket logging: '+str(err))

    def create_encryption(self):
        """
        Set encryption on a bucket
        :return:
        """
        try:
            response = self.client.get_bucket_encryption(
                Bucket=self.bucket_name
            )
            if self.debug:
                print(response)
        except ClientError as err:
            if self.debug:
                print('no bucket encryption:'+str(err))

            response = self.client.put_bucket_encryption(
                Bucket=self.bucket_name,
                ServerSideEncryptionConfiguration={
                    'Rules': [
                        {
                            'ApplyServerSideEncryptionByDefault': {
                                'SSEAlgorithm': 'AES256'
                            }
                        }
                    ]
                }
            )

            if self.debug:
                print(response)


    def add_lifecycle_policy(self):
        """
        Add a lifecycle policy to a bucket
        :return:
        """
        try:
            response = self.client.put_bucket_lifecycle_configuration(
                Bucket=self.bucket_name,
                LifecycleConfiguration={
                    'Rules': [
                        {
                            'ID': 'MoveToStandardIa',
                            'Filter': {
                                'Prefix': ''
                            },
                            'Status': 'Enabled',
                            'Transitions': [
                                {
                                    'Days': self.days_to_standard_ia,
                                    'StorageClass': 'STANDARD_IA'
                                }
                            ]
                        },
                        {
                            'ID': 'MoveToGlacier',
                            'Filter':{
                                'Prefix':''
                            },
                            'Status': 'Enabled',
                            'Transitions': [
                                {
                                    'Days': self.days_to_glacier,
                                    'StorageClass': 'GLACIER'
                                }
                            ]

                        }
                    ]
                }
            )

            if self.debug:
                print(response)
        except ClientError as err:
            print('Error adding bucket lifecycle policy: ' + str(err))


    def add_bucket_policy(self):
        """
        Add a bucket policy
        :return:
        """

        try:

            if self.debug:
                print('##############################')
                print('policy is: '+str(str(self.bucket_policy).replace('"', '\''))+lineno())
                print('##############################')


            if self.bucket_policy_path:
                response = self.client.put_bucket_policy(
                    Bucket=self.bucket_name,
                    Policy=json.loads(json.dumps(self.bucket_policy))
                )

                if self.debug:
                    print(response)

            elif self.bucket_policy and (not self.bucket_policy_path and len(self.bucket_policy_principals) < 1):

                temp_policy_string = self.bucket_policy.replace("\n", '').replace('\'', '"')

                if self.debug:
                    print('temp_policy_string: '+str(temp_policy_string))

                dict_policy = json.loads(temp_policy_string)

                if self.debug:
                    #print('policy as json: '+str(temp_policy))
                    print('policy as dict: '+str(json.loads(temp_policy_string)))
                    print('dict type: '+str(type(dict_policy)))

                response = self.client.put_bucket_policy(
                    Bucket=self.bucket_name,
                    Policy=json.dumps(dict_policy).strip()
                )

                if self.debug:
                    print(response)

            else:
                response = self.client.put_bucket_policy(Bucket=self.bucket_name, Policy=json.dumps(self.bucket_policy))
                if self.debug:
                    print(response)

        except ClientError as err:
            print('Error adding bucket policy: ' + str(err))


    def create_bucket_policy(self):
        """
        Create S3 Bucket Policy
        :return:
        """
        data = {}
        data['Version'] = "2012-10-17"
        data['Statement'] = []
        data['Statement'].append({})
        data['Statement'][0]["Sid"] = "AllowRoot"
        data['Statement'][0]["Effect"] = "Allow"
        data['Statement'][0]["Principal"] = {}
        data['Statement'][0]["Principal"]["AWS"] = []
        data['Statement'][0]["Action"] = []
        data['Statement'][0]["Action"].append("s3:*")
        data['Statement'][0]["Resource"] = []
        data['Statement'][0]["Resource"].append("arn:aws:s3:::"+str(self.bucket_name)+"/*")
        data['Statement'][0]["Resource"].append("arn:aws:s3:::"+str(self.bucket_name))


        for principal in self.bucket_policy_principals:
            data["Statement"][0]["Principal"]["AWS"].append(principal)

        self.bucket_policy = data

    def create_tags(self):
        """
        Create tags on S3 bucket
        :return:
        """
        if self.debug:
            print('create tags'+lineno())

        while not self.check_bucket():
            print('Waiting for bucket to be created')
            print('Sleeping for 5 minutes')
            time.sleep(300)


        if self.debug:
            print('tags to check: '+str(self.tags))

        if self.tags:

            try:

                if self.debug:
                    print('new tag info: '+str(self.tags)+lineno())

                response = self.client.put_bucket_tagging(
                    Bucket=str(self.bucket_name),
                    Tagging={
                        'TagSet': self.tags
                    }
                )

                if self.debug:
                    print('response: '+str(response))

            except ClientError as err:
                print('Error creating tags: '+str(err))

        else:
            print('Create the bucket before trying to add tags')
            sys.exit(1)


    def create_bucket(self):
        """
        Create an S3 Bucket
        :return:
        """
        if self.debug:
            print('create bucket'+lineno())
            print('bucket name: '+str(self.bucket_name))

        if self.check_bucket():
            # Bucket already exists
            return
        else:
            # Create bucket

            try:
                if self.debug:
                    print('Creating bucket')
                response = self.client.create_bucket(Bucket=str(self.bucket_name))

                if self.debug:
                    print('response: '+str(response)+lineno())

            except ClientError as err:
                print('Error: '+str(err))
                sys.exit(1)


    def check_bucket(self):
        """
        Check if a bucket exists
        :return:
        """
        try:
            self.resource.meta.client.head_bucket(Bucket=self.bucket_name)
            if self.debug:
                print("Bucket Already Exists!")

            return True
        except botocore.exceptions.ClientError as err:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(err.response['Error']['Code'])
            if error_code == 403:
                if self.debug:
                    print("Bucket Already Exists - Private Bucket. Forbidden Access!")
                return True
            elif error_code == 404:
                if self.debug:
                    print("Bucket Does Not Exist!")
                return False



#!/usr/bin/env python

from __future__ import absolute_import, division, print_function
from setuptools import setup, find_packages
import sys


DESCRIPTION = ("Creates S3 bucket, tags, policy and lifecycle policy via boto3.")
LONG_DESCRIPTION = open('README.rst').read()
VERSION = '0.0.16'

setup_requires = (
    ['pytest-runner'] if any(x in sys.argv for x in ('pytest', 'test', 'ptr')) else []
)

setup(
    name='aws_bucket_creator',
    version=VERSION,
    description=DESCRIPTION,
    url='https://github.com/rubelw/aws_bucket_creator',
    author='Will Rubel',
    author_email='willrubel@gmail.com',
    long_description=LONG_DESCRIPTION,
    platforms=["any"],
    packages=find_packages(),
    include_package_data=True,
    setup_requires=setup_requires,
    tests_require=['pytest','mock'],
    test_suite="aws_bucket_creator.tests",
    install_requires=[
        "boto3>=1.4.3",
        "requests>=2.18",
        "Click>=6.7",
        "configparser>=3.5.0",
        "future>=0.16.0",
        "six>=1.11.0",
        "pip"
    ],
    keywords=['s3', 'bucket', 'creator'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    entry_points="""
        [console_scripts]
        bucket-creator=aws_bucket_creator.command:cli
    """
)